/*
 Copyright (C) 2011,2012 Antonio Espinoza
 This program is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#include <sys/ptrace.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <unistd.h>
#include <sys/user.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/syscall.h>
#include <string.h>
void printCregs(char* c ,struct user_regs_struct regs, FILE *fp);
void copyRegs(struct user_regs_struct *r1, struct user_regs_struct *r2);
void printRegs(struct user_regs_struct regs, FILE *fp);
void whatChanged(struct user_regs_struct r1, struct user_regs_struct r2,char*cg);
int main(int arg, char *argv[])
{  
    pid_t child;
    char st[20];
    char *s = &st[0];
    const int long_size = sizeof(long);
    child = fork();
    FILE *fp;
    fp = fopen("interceptor.out", "w");
    if (fp == NULL)
        exit(1);
    strcpy(st,"./");
    strcat(st,argv[1]);
    if(child == 0) 
    {
        ptrace(PTRACE_TRACEME, 0, NULL, NULL);
        //execl("./dummy", "dummy", NULL);
        execl(s,argv[1], NULL);
    }
    else 
    {
        int status;
        char changed [26];//a value for every register that changed
        struct user_regs_struct regs;
        struct user_regs_struct oldregs;
        wait(&status);
        ptrace(PTRACE_GETREGS,child, NULL, &regs);
        ptrace(PTRACE_SINGLESTEP, child,NULL, NULL);
        fprintf(fp, "[%lx]",regs.eip);
        copyRegs(&oldregs,&regs);
        printRegs(regs,fp);
        while(1) 
        {
            wait(&status);
            if(WIFEXITED(status))
                break;
            ptrace(PTRACE_GETREGS,child, NULL, &regs);
            fprintf(fp, "[%lx]",regs.eip);
            whatChanged(regs,oldregs,changed); 
            printCregs(changed,regs,fp);
            //printf("%s",changed);
            copyRegs(&oldregs,&regs);
            ptrace(PTRACE_SINGLESTEP, child,NULL, NULL);

        }
    }
    return 0;
}
/*
 *INPUT: Two user_regs_struct pointers
 *OUTPUT: 0 if the values in all registers are the same
 *        1 if any registers differ 
 * */
void whatChanged(struct user_regs_struct r1, struct user_regs_struct r2,char*cg)
{
  int x=0;
  //clear out the garbage in the array 
  for(x=0;x<27;x++)
    cg[x]='\0';
  x = 0;//start at beginning of array
        //and only add chars as they come by refrencing cg[x++]
  //if (r1.r15 !=r2.r15)
  //  cg[x++]='a';
  //if (r1.r14 !=r2.r14)
  //  cg[x++]='b';
  //if (r1.r13 !=r2.r13)
  //  cg[x++]='c';
  //if (r1.r12 !=r2.r12)
  //  cg[x++]='d';
  if (r1.ebp !=r2.ebp)
    cg[x++]='z';
  if (r1.ebx !=r2.ebx)
    cg[x++]='e';
  //if (r1.r11 !=r2.r11)
  //  cg[x++]='f';
  //if (r1.r10 !=r2.r10)
  //  cg[x++]='g';
  //if (r1.r9 !=r2.r9)
  //  cg[x++]='h';
  //if (r1.r8 !=r2.r8)
  //  cg[x++]='i';
  if (r1.eax !=r2.eax)
    cg[x++]='j';
  if (r1.ecx !=r2.ecx)
    cg[x++]='k';
  if (r1.edx !=r2.edx)
    cg[x++]='l';
  if (r1.esi !=r2.esi)
    cg[x++]='m';
  if (r1.edi !=r2.edi)
    cg[x++]='n';
  if (r1.orig_eax !=r2.orig_eax)
    cg[x++]='o';
  if (r1.xcs !=r2.xcs)
    cg[x++]='p';
  if (r1.eflags !=r2.eflags)
    cg[x++]='q';
  if (r1.esp !=r2.esp)
    cg[x++]='r';
  if (r1.xss !=r2.xss)
    cg[x++]='s';
  //if (r1.fs_base !=r2.fs_base)
  //  cg[x++]='t';
  //if (r1.gs_base !=r2.gs_base)
  //  cg[x++]='u';
  if (r1.xds !=r2.xds)
    cg[x++]='v';
  if (r1.xes !=r2.xes)
    cg[x++]='w';
  if (r1.xfs !=r2.xfs)
    cg[x++]='x';
  if (r1.xgs !=r2.xgs)
    cg[x]='y';
}
/*
 *This function should take two pointers to the structure:
 *user_regs_struct 
 *and put the contents of the second into the first
 */
void copyRegs(struct user_regs_struct *r1, struct user_regs_struct *r2)
{
 //(*r1).r15 = (*r2).r15;
 //(*r1).r14 = (*r2).r14;
 //(*r1).r13 = (*r2).r13;
 //(*r1).r12 = (*r2).r12;
 (*r1).ebp = (*r2).ebp;
 (*r1).ebx = (*r2).ebx;
 //(*r1).r11 = (*r2).r11;
 //(*r1).r10 = (*r2).r10;
 //(*r1).r9 = (*r2).r9;
 //(*r1).r8 = (*r2).r8;
 (*r1).eax = (*r2).eax;
 (*r1).ecx = (*r2).ecx;
 (*r1).edx = (*r2).edx;
 (*r1).esi = (*r2).esi;
 (*r1).edi = (*r2).edi;
 (*r1).orig_eax = (*r2).orig_eax;
 (*r1).eip = (*r2).eip;
 (*r1).xcs = (*r2).xcs;
 (*r1).eflags = (*r2).eflags;
 (*r1).esp = (*r2).esp;
 (*r1).xss = (*r2).xss;
 //(*r1).fs_base = (*r2).fs_base;
 //(*r1).gs_base = (*r2).gs_base;
 (*r1).xds = (*r2).xds;
 (*r1).xes = (*r2).xes;
 (*r1).xfs = (*r2).xfs;
 (*r1).xgs = (*r2).xgs;

}
void printRegs(struct user_regs_struct regs, FILE *fp)
{
  /*
  printf("r15=[%lx]",regs.r15);
  printf("r14=[%lx]",regs.r14);
  printf("r13=[%lx]",regs.r13);
  printf("r12=[%lx]",regs.r12);
  */
  fprintf(fp, "ebp=[%lx]",regs.ebp);
  fprintf(fp, "ebx=[%lx]",regs.ebx);
  //printf("r11=[%lx]",regs.r11);
  //printf("r10=[%lx]",regs.r10);
  //printf("r9=[%lx]",regs.r9);
  //printf("r8=[%lx]",regs.r8);
  fprintf(fp, "eax=[%lx]",regs.eax);
  fprintf(fp, "ecx=[%lx]",regs.ecx);
  fprintf(fp, "edx=[%lx]",regs.edx);
  fprintf(fp, "esi=[%lx]",regs.esi);
  fprintf(fp, "edi=[%lx]",regs.edi);
  fprintf(fp, "orig_eax=[%lx]",regs.orig_eax);
  fprintf(fp, "eip=[%lx]",regs.eip);
  fprintf(fp, "xcs=[%lx]",regs.xcs);
  fprintf(fp, "eflags=[%lx]",regs.eflags);
  fprintf(fp, "esp=[%lx]",regs.esp);
  fprintf(fp, "xss=[%lx]",regs.xss);
  //printf("fs_base=[%lx]",regs.fs_base);
  //printf("gs_base=[%lx]",regs.gs_base);
  fprintf(fp, "xds=[%lx]",regs.xds);
  fprintf(fp, "xes=[%lx]",regs.xes);
  fprintf(fp, "xfs=[%lx]",regs.xfs);
  fprintf(fp, "xgs=[%lx]\n",regs.xgs);
}
void printCregs(char* c, struct user_regs_struct regs, FILE *fp)
{
  int x = 0;
  for(x=0;x<28;x++)
  {
    /*
    if((c[x])=='a')
      printf("r15=[%lx]",regs.r15);
    if((c[x])=='b')
      printf("r14=[%lx]",regs.r14);
    if((c[x])=='c')
      printf("r13=[%lx]",regs.r13);
    if((c[x])=='d')
      printf("r12=[%lx]",regs.r12);
    */
    if((c[x])=='z')
      fprintf(fp, "ebp=[%lx]",regs.ebp);
    if((c[x])=='e')
      fprintf(fp, "ebx=[%lx]",regs.ebx);
    //if((c[x])=='f')
    //  printf("r11=[%lx]",regs.r11);
    //if((c[x])=='g')
    //  printf("r10=[%lx]",regs.r10);
    //if((c[x])=='h')
    //  printf("r9=[%lx]",regs.r9);
    //if((c[x])=='i')
    //  printf("r8=[%lx]",regs.r8);
    if((c[x])=='j')
      fprintf(fp, "eax=[%lx]",regs.eax);
    if((c[x])=='k')
      fprintf(fp, "ecx=[%lx]",regs.ecx);
    if((c[x])=='l')
      fprintf(fp, "edx=[%lx]",regs.edx);
    if((c[x])=='m')
      fprintf(fp, "esi=[%lx]",regs.esi);
    if((c[x])=='n')
      fprintf(fp, "edi=[%lx]",regs.edi);
    if((c[x])=='o')
      fprintf(fp, "origeax=[%lx]",regs.orig_eax);
    if((c[x])=='p')
      fprintf(fp, "xcs=[%lx]",regs.xcs);
    if((c[x])=='q')
      fprintf(fp, "eflags=[%lx]",regs.eflags);
    if((c[x])=='r')
      fprintf(fp, "esp=[%lx]",regs.esp);
    if((c[x])=='s')
      fprintf(fp, "xss=[%lx]",regs.xss);
    //if((c[x])=='t')
   //   printf("fs_base=[%lx]",regs.fs_base);
    //if((c[x])=='u')
    //  printf("gs_base=[%lx]",regs.gs_base);
    if((c[x])=='v')
      fprintf(fp, "xds=[%lx]",regs.xds);
    if((c[x])=='w')
      fprintf(fp, "xes=[%lx]",regs.xes);
    if((c[x])=='x')
      fprintf(fp, "xfs=[%lx]",regs.xfs);
    if((c[x])=='y')
      fprintf(fp, "xgs=[%lx]",regs.xgs);
    if (c[x]=='\0')
      break;
  }
  fprintf(fp, "\n");
}

