/*
 Copyright (C) 2011,2012 Antonio Espinoza
 This program is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
import java.util.Random;
public class vectFuncts 
{
	public static double twoNorm(float mat [])
	{
		double sqrsum =0;
		for(int i = 0;i<ptraceDIFT.dim;i++)
			sqrsum+=mat[i]*mat[i];
		return Math.sqrt(sqrsum);
	}

	public static double dotprod(float mat1[],float mat2[])
	{
		double dp = 0;
		for(int i=0;i<ptraceDIFT.dim;i++)
			dp += mat1[i]*mat2[i];
		return dp;
	}

	public static double cossim(float mat1[],float mat2[])
	{
		return dotprod(mat1,mat2)/(twoNorm(mat1)*twoNorm(mat2));
	}

	public static double getTaintMagnitude(float mat1[],float mat2[])
	{
		double normMat1 = twoNorm(mat1); 
		double normMat2 = twoNorm(mat2);
		return Math.max(normMat1,normMat2) + Math.min(normMat1,normMat2) *
		       (1-Math.exp(ptraceDIFT.dim*Math.log(Math.abs(cossim(mat1,mat2)))));
		//Making sure the log won't fail due to negative numbers
		//may want to change this to (1 - cossim) round to 0 if <0
		
	}
	public static float [] combineTaintVectors(float mat1[],float mat2[])
	{
		float taintVector [] = new float[ptraceDIFT.dim];
		double tm = getTaintMagnitude(mat1,mat2);
		double scaleV = 0;
		int sum = matSum(mat1);
		int sum2 = matSum(mat2);
		/*Add togeter the vectors and scale by the value returned
		 * from getTaintMagnitude*/
		for (int i=0;i<ptraceDIFT.dim;i++)
			taintVector [i] = mat1[i]+mat2[i];
		if(sum==0||sum2==0)
			return taintVector;
		if(tm > 32) //we can't have more than 32 bits of taint so throttle back
			scaleV = getScaleValue(32,taintVector);
		else
			scaleV = getScaleValue(tm,taintVector);
		for (int i = 0;i<ptraceDIFT.dim;i++)
			taintVector[i]*=scaleV;
		return taintVector;
	}
	public static int matSum(float[] mat1) 
	{
		int sum=0;
		for(int i=0;i<mat1.length;i++)
			sum += mat1[i];
		return sum;
	}
	/*
	 * This should return the value that will need to multiply every 
	 * element of the array by to achieve the desired magnitude.
	 * */
	public static double getScaleValue(double taintMag, float mat[])
	{
		double acc = 0;
		for(int i = 0;i<ptraceDIFT.dim;i++)
			acc += (double)mat[i]*(double)mat[i];
		return Math.sqrt((taintMag*taintMag)/acc);
	}
	public static int getDeltaAndTrim(StringBuffer sb)
	{
		int delta;
		if(sb.charAt(0)=='-')
			delta = -(Integer.parseInt((sb.substring(3, sb.indexOf("("))),16));
		else
			delta = Integer.parseInt((sb.substring(2, sb.indexOf("("))), 16);
		return delta;
	}

	public static void trimExtra(StringBuffer sb)
	{
		sb.delete(0, sb.indexOf("("));
		sb.delete(0,2); //Delete the "(%"
		sb.deleteCharAt(sb.length()-1); //Delete the ")" at the end
	}
	/*
	 * Make sure the correct bits are affected 
	 * http://en.wikibooks.org/wiki/X86_Assembly/X86_Architecture
	 * */
	public static String [] flagsChanged(long o ,long n)
	{
		long chgd = o ^ n;
		boolean c = false;
		StringBuffer sb = new StringBuffer();
		if ((chgd & 0x800) > 0)
		{
			c = true;
			sb.append("OF,");
		}
		if ((chgd & 1) > 0)
		{
			c = true;
			sb.append("CF,");
		}
		if ((chgd & 4) > 0)
		{
			c = true;
			sb.append("PF,");
		}
		if ((chgd & 0x80) > 0)
		{
			c = true;
			sb.append("SF,");
		}
			
		if ((chgd & 0x10) > 0)
		{
			c = true;
			sb.append("AF,");
		}
		if ((chgd & 0x40) > 0)
		{
			sb.append("ZF,");
			c= true;
		}
		if(sb.length()>0)
			sb.deleteCharAt(sb.length()-1);
		if (c)
			return sb.toString().trim().split(",");
		return (new String [0]);
	}
	/*
	 * takes a source of form 0x430(%ecx,%edx,4)
	 * and the original arg and returns the
	 * Offset,Base,Index,Width in a string array in that order
	 * It can take cases where there is no offset or no base
	 * */
	public static String[] getOBIW(StringBuffer src,String arg) 
	{
		String [] retStr = new String[4];
		long OFFSET=0;
		if(arg.startsWith("0x"))
		{
			String soffset = src.substring(0, src.indexOf("("));
			OFFSET = Long.parseLong(soffset,16);
			src.delete(0, src.indexOf("(")+1);
		}
		
		if (src.charAt(0)=='(')
			src.deleteCharAt(0);
		
		String BASE;
		if(src.charAt(0)=='%')
			BASE = src.substring(1,src.indexOf(","));
		else
			BASE = src.substring(0,src.indexOf(","));
		src.delete(0, src.indexOf(",")+1);
		
		String INDEX = src.substring(0,src.indexOf(","));
		INDEX = INDEX.substring(1);
		src.delete(0,src.indexOf(",")+1);
		String swidth = src.substring(0,src.length()-1);
		int WIDTH = Integer.parseInt(swidth);
		retStr[0]=Long.toString(OFFSET);
		retStr[1]=BASE;
		retStr[2]=INDEX;
		retStr[3]=Integer.toString(WIDTH);
		return retStr;
	}
	public static String[] getOBIW(StringBuffer src) 
	{
		String [] retStr = new String[4];
		long OFFSET=0;
		
		if(src.toString().matches("(-??0x\\p{XDigit}+)\\(.*"))//we have an offset 
		{
			if(src.charAt(0)=='-')
			{
				//we are a negative offset
				src.delete(0, 3);
				OFFSET = -(Long.parseLong(src.substring(0, src.indexOf("(")), 16));
				src.delete(0, src.indexOf("("));
//				System.out.println(src);
			}
			else
			{
				src.delete(0, 2);
				OFFSET = (Long.parseLong(src.substring(0, src.indexOf("(")), 16));
				src.delete(0, src.indexOf("("));
//				System.out.println(src);
			}
		}
		
		if (src.charAt(0)=='(')
			src.deleteCharAt(0);
		
		String BASE;
		if(src.charAt(0)=='%')
			BASE = src.substring(1,src.indexOf(","));
		else
			BASE = src.substring(0,src.indexOf(","));
		src.delete(0, src.indexOf(",")+1);
		
		String INDEX = src.substring(0,src.indexOf(","));
		INDEX = INDEX.substring(1);
		src.delete(0,src.indexOf(",")+1);
		String swidth = src.substring(0,src.length()-1);
		int WIDTH = Integer.parseInt(swidth);
		retStr[0]=Long.toString(OFFSET);
		retStr[1]=BASE;
		retStr[2]=INDEX;
		retStr[3]=Integer.toString(WIDTH);
		return retStr;
	}
	public static float[] getRandomTaintVector() 
	{
		Random rdm = new Random();
		double sqrsum = 0;
		float x;
		float [] ntv = new float [ptraceDIFT.dim];
		for (int i =0;i<ptraceDIFT.dim;i++)
		{
			x = (float)rdm.nextGaussian();
			ntv[i]=x;
			sqrsum += Math.pow(x, 2);
		}
		sqrsum = Math.sqrt(sqrsum);
		for(int i =0;i<ptraceDIFT.dim;i++)
			ntv[i]= ((float)(1/sqrsum)*ntv[i])*ptraceDIFT.wordLen;//*32 for a length of 32 bits == size of word
		return ntv;
	}
	
	public static String get32From16(String part)
	{
		if (part.matches("[ds]i"))
			return "e"+part;
		if (part.equals("ch"))
			return "ecx";
		return part;
	}

	public static Long getFinalDest(long O,long B, long I, long W) 
	{
		return I*W+B+O;
	}
	
}
